# -*- coding: utf-8 -*-
#
# fumi deployment tool
# https://github.com/rmed/fumi
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Rafael Medina García <rafamedgar@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function
from deployments import *
import argparse
import os
import sys
import yaml
if sys.version[0] == "3": raw_input=input

__version__ = "0.1.0"

CWD = os.getcwd()
DEP_CONF = os.path.join(CWD, 'fumi.yml')


class Deployment(object):
    """ Configuration parsed from the fumi.yml file """

    def __init__(self, **kwargs):
        # Source information
        self.s_type = kwargs['source-type']
        self.s_path = kwargs['source-path']

        # Pre-deployment commands
        predep = kwargs.get('predep')
        if predep:
            self.pre_local = predep.get('local')
            self.pre_remote = predep.get('remote')
        else:
            self.pre_local = self.pre_remote = None

        # Post-deployment commands
        postdep = kwargs.get('postdep')
        if postdep:
            self.post_local = postdep.get('local')
            self.post_remote = postdep.get('remote')
        else:
            self.post_local = self.post_remote = None

        # Destination host information
        self.host = kwargs['host']
        self.user = kwargs['user']
        self.d_path = kwargs['deploy-path']
        self.h_tmp = kwargs.get('host-tmp')

        # Optional information
        self.keep = kwargs.get('keep-max')
        self.default = kwargs.get('default')


def deploy(configuration):
    """ Deploy with given configuration. """
    content = read_yaml()
    if not content:
        sys.exit("There is no fumi.yml file")

    if not configuration:
        default = None

        for k in content.keys():
            if content[k].get("default"):
                default = k
                print("Using default configuration '%s'" % default)
                break

        if not default:
            if len(content.keys()) == 0:
                sys.exit("There are no configurations")

            # Ask for default
            print("I found the following configurations:")

            for k in content.keys():
                print("-", k)

            default = raw_input("Which one do you want to set as default?: ")

            if default in content.keys():
                content[default]["default"] = True
                write_yaml(content)

            else:
                sys.exit("That configuration does not exist!")

        configuration = default

    elif configuration not in content.keys():
        sys.exit("Configuration '%s' does not exist" % configuration)

    # Build deployment object
    dep = Deployment(**content[configuration])

    # Local copy deployment
    if dep.s_type == "local":
        deploy_local(dep)

    elif dep.s_type == "git":
        deploy_git(dep)


def list_configs():
    """ List the configurations present in the fumi.yml file. """
    content = read_yaml()
    if not content:
        sys.exit("There is no fumi.yml file")

    for conf in content.keys():
        print(conf)

def new_config(name):
    """ Create new basic configuration in fumi.yml file. """
    content = read_yaml()
    if not content:
        content = {}

    if name in content.keys():
        sys.exit("Configuration '%s' already exists" % name)

    content[name] = {
        "source-type": "",
        "source-path": "",

        "predep": {
            "local": None,
            "remote": None
        },
        "postdep": {
            "local": None,
            "remote": None
        },

        "host": "",
        "user": "",
        "deploy-path": "",
    }

    write_yaml(content)

    print("Created new blank configuration '%s'" % name)

def remove_config(name):
    """ Remove a configuration from the fumi.yml file. """
    content = read_yaml()
    if not content:
        sys.exit("There is no fumi.yml file")

    if name not in content.keys():
        sys.exit("Configuration %s does not exist" % name)

    del content[name]

    write_yaml(content)

    print("Removed configuration '%s'" % name)

def read_yaml():
    """ Reads the fumi.yml file and returns the parsed information
        in a dict() object.

        If there is no file, then returns None
    """
    if os.path.isfile(DEP_CONF):
        with open(DEP_CONF, 'r') as fumi_yml:
            try:
                content = yaml.load(fumi_yml)

            except (yaml.YAMLError, e):
                sys.exit("Error in deployment file:", e)

        return content

    return None

def write_yaml(content):
    """ Overwrites the content of the fumi.yml file. """
    with open(DEP_CONF, 'w') as fumi_yml:
        try:
            yaml.dump(content, fumi_yml, default_flow_style=False)

        except (yaml.YAMLError, e):
            sys.exit("Error writing yaml to deployment file:", e)

def init_parser():
    """ Initialize the arguments parser. """
    parser = argparse.ArgumentParser(
        description="Simple deployment tool")

    parser.add_argument('--version', action='version',
        version='%(prog)s ' + __version__)

    subparsers = parser.add_subparsers(title="commands")

    # deploy
    parser_deploy = subparsers.add_parser("deploy",
        help="deploy with given configuration")
    parser_deploy.add_argument("configuration", nargs="?",
        help="configuration to use")

    # list
    parser_list = subparsers.add_parser("list",
        help="list all the available deployment configurations")

    # new
    parser_new = subparsers.add_parser("new",
        help="create new deployment configuration")
    parser_new.add_argument("name", help="name for the configuration")

    # remove
    parser_remove = subparsers.add_parser("remove",
        help="remove a configuration from the deployment file")
    parser_remove.add_argument("name", help="name of the configuration")

    return parser

def parse_action(action, parsed):
    """ Parse the action to execute. """
    if action == 'deploy':
        deploy(parsed.configuration)

    elif action == 'list':
        list_configs()

    elif action == 'new':
        new_config(parsed.name)

    elif action == 'remove':
        remove_config(parsed.name)

    else:
        parser.print_help()


def main():
    parser = init_parser()
    args = parser.parse_args()

    if len(sys.argv) == 1:
        # No argument provided
        parser.print_help()
        return

    parse_action(sys.argv[1], args)



# -*- coding: utf-8 -*-
#
# fumi deployment tool
# https://github.com/rmed/fumi
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Rafael Medina García <rafamedgar@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function
from blessings import Terminal
import datetime
import os
import paramiko
import scp
import subprocess
import sys
import tarfile


term = Terminal()
ssh = paramiko.SSHClient()
ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())


def deploy_local(deployment):
    """ Local based deployment. """
    ssh.connect(deployment.host, username=deployment.user)

    _predep(deployment.pre_local, deployment.pre_remote)


    # Compress source to temporary directory
    timestamp = datetime.datetime.utcnow().strftime("%Y%m%d%H%M%S")
    comp_file = timestamp + ".tar.gz"
    tmp_local = os.path.join("/tmp", comp_file)

    print(term.bold_cyan("\n# Compressing source to /tmp/%s\n" % comp_file))

    with tarfile.open(tmp_local, "w:gz") as tar:
        tar.add(deployment.s_path, arcname=timestamp)

    print(term.bold_green("Done!"))


    # Upload compressed source
    print(term.bold_cyan("\n# Uploading %s...\n" % comp_file))

    uload = scp.SCPClient(ssh.get_transport())
    uload_tmp = deployment.h_tmp or "/tmp"
    uload_path = os.path.join(uload_tmp, comp_file)

    try:
        uload.put(tmp_local, uload_path)

    except scp.SCPException as e:
        sys.exit(term.bold_red("Error uploading to server: %s" % e))

    print(term.bold_green("Done!"))


    # Uncompress source to deploy_path/rev
    print(term.bold_cyan("\n# Uncompressing remote file...\n"))

    rev_path = os.path.join(deployment.d_path, "rev")
    untar = "tar -C %s -zxvf %s" % (rev_path, uload_path)

    stdin, stdout, stderr = ssh.exec_command(untar)
    status = stdout.channel.recv_exit_status()

    if status == 127:
        sys.exit(term.bold_red("Error: tar command not found in remote host"))
    elif status == 1:
        print(*stderr.readlines())
        sys.exit(term.bold_red("Error: some files differ"))
    elif status == 2:
        print(*stderr.readlines())
        sys.exit(term.bold_red("Fatal error when extracting remote file"))

    print(term.bold_green("Done!"))


    # Link directory
    _symlink(deployment.d_path, rev_path, timestamp)


    # Run post-deployment commands
    _postdep(deployment.post_local, deployment.post_remote,
        os.path.join(deployment.d_path, "current"))


    # Clean revisions
    if deployment.keep:
        _clean_revisions(deployment.keep, rev_path)


    # Cleanup temporary files
    print(term.bold_cyan("\n# Cleaning temporary files...\n"))

    subprocess.call(["rm", tmp_local])
    stdin, stdout, stderr = ssh.exec_command("rm %s" % uload_path)

    print(term.bold_green("Done!"))


    print(term.bold_green("\nDeployment complete!"))

def deploy_git(deployment):
    """ Git based deployment. """
    ssh.connect(deployment.host, username=deployment.user)

    # Run pre-deployment commands
    _predep(deployment.pre_local, deployment.pre_remote)


    # Clone source
    print(term.bold_cyan("\n# Cloning repository...\n"))

    timestamp = datetime.datetime.utcnow().strftime("%Y%m%d%H%M%S")
    rev_path = os.path.join(deployment.d_path, "rev")
    current_rev = os.path.join(deployment.d_path, "rev", timestamp)
    clone = "git clone %s %s" % (deployment.s_path, current_rev)

    stdin, stdout, stderr = ssh.exec_command(clone)
    status = stdout.channel.recv_exit_status()

    if status == 127:
        sys.exit(term.bold_red("git command not found in remote server"))
    # TODO: check git exit codes

    print(term.bold_green("Done!"))


    # Link directory
    _symlink(deployment.d_path, rev_path, timestamp)


    # Run post-deployment commands
    _postdep(deployment.post_local, deployment.post_remote,
        os.path.join(deployment.d_path, "current"))


    # Clean revisions
    if deployment.keep:
        _clean_revisions(deployment.keep, rev_path)


    print(term.bold_green("\nDeployment complete!"))


def _clean_revisions(keep, rev_path):
    """ Remove old revisions from the remote server.

        Only the most recent revisions will be kept, to a maximum defined
        by the keep argument.
    """
    print(term.bold_cyan("\n# Checking old revisions...\n"))

    stdin, stdout, stderr = ssh.exec_command("ls %s" % rev_path)
    status = stdout.channel.recv_exit_status()

    if status == 1 or status == 2:
        print(*stderr.readlines())
        sys.exit(term.bold_red("Error obtaining list of revisions"))

    revisions = stdout.readlines()

    if len(revisions) > keep:
        old_revisions = []

        while len(revisions) > keep:
            old_revisions.append(revisions.pop(0))

        for r in old_revisions:
            print(term.bold_magenta("Removing revision %s" % r.strip()))
            rm_old = "rm -rf %s" % os.path.join(rev_path, r.strip())
            stdin, stdout, stderr = ssh.exec_command(rm_old)

    print(term.bold_green("\nDone!"))

def _postdep(local, remote, link_path):
    """ Execute post-deployment commands (both local and remote).

        Note that remote post-deployment commands are executed in the
        directory of the revision that has been deployed
        (deploy_path/current).
    """
    if local:
        for cmd in local:
            print(term.bold_cyan("\n# Running local command: %s\n" % cmd))
            subprocess.Popen(cmd, shell=True).wait()

    # Executed relative to deploy_path/current
    if remote:
        for cmd in remote:
            print(term.bold_cyan("\n# Running remote command: %s\n" % cmd))
            cd_command = "cd %s; %s" % (link_path, cmd)
            stdin, stdout, stderr = ssh.exec_command(cd_command)
            # TODO: better handling for SSH return codes?
            # print(stdout.channel.recv_exit_status())
            print(*stdout.readlines())

def _predep(local, remote):
    """ Execute pre-deployment commands (both local and remote).

        Note that remote pre-deployment commands are usually executed
        in the users directory (~/) when possible.
    """
    if local:
        for cmd in local:
            print(term.bold_cyan("\n# Running local command: %s\n" % cmd))
            subprocess.Popen(cmd, shell=True).wait()

    # Executed relative to user directory
    if remote:
        for cmd in remote:
            print(term.bold_cyan("\n# Running remote command: %s\n" % cmd))
            stdin, stdout, stderr = ssh.exec_command(cmd)
            # TODO: better handling for SSH return codes?
            # print(stdout.channel.recv_exit_status())
            print(*stdout.readlines())

def _symlink(d_path, rev_path, timestamp):
    """ Symlink the deployed revision to the deploy_path/current
        directory.
    """
    print(term.bold_cyan("\n# Linking directory...\n"))

    link_path = os.path.join(d_path, "current")
    current_rev = os.path.join(rev_path, timestamp)
    ln = "ln -sfn %s %s" % (current_rev, link_path)

    stdin, stdout, stderr = ssh.exec_command(ln)
    # status = stdout.channel.recv_exit_status()

    print(term.bold_green("Done!"))
